(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:Title: DRAKE: Dark matter Relic Abundance beyond Kinetic Equilibrium

:Context: DRAKE`

:Summary:

   DRAKE is a numerical precision tool for predicting the dark matter relic abundance
   also in situations where the standard assumption of kinetic equilibrium
   during the freeze-out process may not be satisfied. DRAKE comes with a
   set of three dedicated Boltzmann equation solvers that implement, respectively,
   the usual equation for the dark matter number density, fluid-like equations
   that couple the evolution of number density and velocity dispersion, and
   a full numerical evolution of the phase-space distribution function.

:Authors:

  Andrzej Hryczuk (NCBJ)
  Tobias Binder (Kavli IMPU)
  Michael Gustafsson (Goettingen U.)
  Torsten Bringmann (Oslo U.)

  The original phase-space density solver (used in arXiv:1706.07433) was 
  developed and implemented in Matlab by: Michael Gustafsson


:History:

   v1.0 - first release

:Reference:

  [arXiv:2103.XXXXX], and
  Phys.Rev.D 96 (2017) 11, 115010 [arXiv:1706.07433] for underlying physics.


:Copyright:

 \[Copyright]2021
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
$ContextPath = Prepend[$ContextPath, "DRAKE`Private`"];

(* print intro *)
Print["Welcome to DRAKE v1.0"];


BeginPackage["DRAKE`"];

(************************** function  descriptions ****************************)
nBE::usage = "Routine calculating relic density under the assumption of kinetic equilibrium, see nBE.wl in src directory. Needs function svx[x]. Output: Oh2nBE, tYnBE.";
cBE::usage = "Routine calculating relic density with coupled Boltzmann equations for number density and temperature, see cBE.wl in src directory. Needs functions svx[x],sv2x[x],gam[x]. Output: Oh2cBE, tYcBE, tycBE.";
fBE::usage = "Routine calculating relic density on phase-space density level, see fBE.wl in src directory. Needs function gam[x] and table tpregsvtheta. Output: Oh2fBE, tYfBE, tyfBE, tfDM.";
PrepANN::usage = "Routine tabulating and interpolating the <sv>, see rates.wl in src directory. Output: svx, dsvx.";
PrepANN2::usage = "Routine tabulating and interpolating the <sv>_2, see rates.wl in src directory. Output: sv2x, dsv2x.";
PrepSCATT::usage = "Routine tabulating the scattering rate, gamma(x), see rates.wl in src directory. Output: gam.";
PrepareANNtheta::usage = "Routine calculating and tabulating <sv>_theta, see rates.wl in src directory";
PrepANNtheta::usge = "Combines call of PrepareANNtheta and RegArrayGen, see rates.wl in src directory";
RegArrayGen::usage = "Routine generating a uniform dense table out of initial non-uniform <sv>_theta tabulation, see rates.wl in src directory. Output: tpregsvtheta.";
tpregsvtheta::usage = "Output of RegArrayGen. Uniform grid table of <sv>_theta.";
save::usage = "Routine streamlining the saving of the output";
Oh2nBE::usage = "Result for the relic density with nBE";
Oh2cBE::usage = "Result for the relic density with cBE";
Oh2fBE::usage = "Result for the relic density with fBE";
tYnBE::usage = "Result for the {x,Y} table with nBE";
tYcBE::usage = "Result for the {x,Y} table with cBE";
tycBE::usage = "Result for the {x,y} table with cBE";
tYfBE::usage = "Result for the {x,Y} table with fBE";
tyfBE::usage = "Result for the {x,y} table with fBE";
tfDM::usage = "Result for the {x,{q^2 fDM(x,q)}} table with fBE at three different times (x initial, x at A to B variable transition, and x final) and fDM is normalized. Adapt fBE.wl (in source directoy) for obtaining other output.";
feq::usage = "Function for equilibrium f(x,p), see auxiliary.wl in src directory.";
Yeq::usage = "Function for equilibrium Y(x), see auxiliary.wl in src directory.";
yeq::usage = "Function for equilibrium y(x), see auxiliary.wl in src directory.";
neq::usage = "Function for equilibrium n(x), see auxiliary.wl in src directory.";
gam::usage = "Function for momentum transfer rate gamma(x)";
svx::usage = "Function for <sv>(x)";
dsvx::usage = "Function for first derivative d/dx <sv>(x)";
sv2x::usage = "Function for <sv>_2(x)";
dsv2x::usage = "Function for first derivative d/dx <sv>_2(x)";
Tabulate1Dloglog::usage = "Routine performing an iterative tabulation in log-log";
TAsvx::usage = "Function calculating the thermal averaged cross section <sv>(x) at point x";
TAsv2x::usage = "Function calculating the 2nd moment of the thermal averaged cross section <sv>_2(x) at point x";
TAsvxnonrel::usage = "Function calculating the thermal averaged cross section <sv>(x) at point x - deeply non relativistic version";
TAsv2xnonrel::usage = "Function calculating the 2nd moment of the thermal averaged cross section <sv>_2(x)  at point x - deeply non relativistic version";
H::usage = "The Hubble rate; H[T], see dof.wl in src directory.";
s::usage = "Entropy density; s[T], see dof.wl in src directory.";
gtilde::usage = "g~ = T/3 h_eff'[T]/h_eff[T], where h_eff[T] is the # of entropy degrees of freedom, see dof.wl in src directory.";



(* set appropriate paths *)
SetDirectory[AbsoluteFileName["."]];
AppendTo[$Path, "./src/"];

(* Include all the package routines in a Private context *)
Begin["`Private`"];

(**************************** include src files *******************************)
Get["auxiliary.wl"];     (* auxiliary routines; needs to be included first *)
Get["dof.wl"];           (* import entropy and energy degrees of freedom *)
Get["thermalavg.wl"];    (* routines for thermal averages *)
Get["rates.wl"];         (* routines for rate calculations *)
Get["nBE.wl"];           (* nBE - BE solver for only number density *)
Get["cBE.wl"];           (* cBE - BE solver for coupled n_DM and T_DM *)
Get["fBE.wl"];           (* fBE - BE solver for f(p)_DM *)
Get["compiled.wl"];      (* compiled functions used in cBE and fBE *)

End[]; (* end context *)

EndPackage[ ];  (* end the package context *)
