(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
C-compiled functions

Version: 15-01-2021
Contact: Tobias Binder (Kavli IPMU)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

GetRelTempCorrection=Compile[{{X,_Real,0},{ExpDbk2,_Real,0},{derbk2Dbk2,_Real,0}},Module[{lmin,lmax,l,dl,fDM,bins,w34,w34prime},

    (* Computes w34 and its derivative. Accurate for x>5. *)

    bins        =   220;
    lmin        =   10^(-4)Sqrt[2X];
    lmax        =   8.*Sqrt[2X];
    dl          =   (lmax-lmin)/(bins-1.);
    l           =   Table[lmin+(k-1.) dl,{k,bins}];
    fDM         =   Exp[-(Sqrt[X^2+l^2]-X)]*ExpDbk2;
    w34         =   (1./3.)*(X^2)^(-1)*dl*((l^6/((X^2+l^2)^(3/2))).fDM);
    w34prime    =   -(2./X+derbk2Dbk2)*w34-(X)^(-1)*dl*((l^6/((X^2+l^2)^(5/2))).fDM)-(3X)^(-1)*dl*((l^6/((X^2+l^2)^(2))).fDM);

    {w34,w34prime}

], RuntimeOptions->"Speed", CompilationTarget->"C"];

(*----------------------------------------------------------------------------*)

GetDerivative=Compile[{{Dim, _Integer},{Order, _Integer}, {Power, _Integer}}, Module[ {i,der},

    (* Returns first (power=1) or second derivative (power=2) to fourth order accuracy.
    Boundary conditions are reflection symmetry at the origin and zero phase-space density at infinity.
    *)

    der                         =   ConstantArray[0.,{Dim,Dim}];

    If[Power==1,
        For[i=3,i<Dim-1,i++,
            der[[i,i-2]]        =   1./12.;
            der[[i,i-1]]        =   -2./3.;
            der[[i,i+1]]        =   2./3.;
            der[[i,i+2]]        =   -1./12.;
        ];

        (*boundary condition origin*)

        der[[2,1]]              =   -2./3.;
        der[[2,2]]              =   1./12.;
        der[[2,3]]              =   2./3.;
        der[[2,4]]              =   -1./12.;

        (*boundary condition infinity*)

        der[[Dim-1,Dim-1-2]]    =   1./12.;
        der[[Dim-1,Dim-1-1]]    =   -2./3.;
        der[[Dim-1,Dim-1+1]]    =   2./3.;
        der[[Dim,Dim-2]]        =   1./12.;
        der[[Dim,Dim-1]]        =   -2./3.;
        der[[Dim,Dim]]          =   0.;

    ];

    If[Power==2,
        For[i=3,i<Dim-1,i++,
            der[[i,i-2]]        =   -1./12.;
            der[[i,i-1]]        =   4./3.;
            der[[i,i]]          =   -5./2.;
            der[[i,i+1]]        =   4./3.;
            der[[i,i+2]]        =   -1./12.;
        ];

        (*origin*)

        der[[1,1]]              =   -5./2.;
        der[[1,1+1]]            =   8./3.;
        der[[1,1+2]]            =   -2./12.;
        der[[2,1]]              =   4./3.;
        der[[2,1+1]]            =   -5./2.-1./12.;
        der[[2,1+2]]            =   4./3.;
        der[[2,1+3]]            =   -1./12.;

        (*infinity*)

        der[[Dim-1,Dim-1-2]]    =   -1./12.;
        der[[Dim-1,Dim-1-1]]    =   4./3.;
        der[[Dim-1,Dim-1]]      =   -5./2.;
        der[[Dim-1,Dim-1+1]]    =   4./3.;
        der[[Dim,Dim-2]]        =   -1./12.;
        der[[Dim,Dim-1]]        =   4./3.;
        der[[Dim,Dim]]          =   -5./2.;

    ];

    der

], RuntimeOptions->"Speed", CompilationTarget->"C"];

(*----------------------------------------------------------------------------*)

GetAnnMatrixTab=Compile[{{pref,_Real,0},{dim,_Integer,0},{scalar,_Real,0},{vector,_Real,1},{arr,_Real,2},{lt,_Integer,0},{lpmin,_Real,0},{lpmax,_Real,0}},
Module[{i,j,matrix,p,pi,pj,crossx=0.,nf,n,mf,m,fi=0.},

    p                       =   Log[scalar vector];(* transfor q into Log[p]*)
    matrix                  =   ConstantArray[0.,{dim,dim}];

    Do[
        pi                  =   p[[i]];
        nf                  =   (lt-1)(pi-lpmin)/(lpmax-lpmin);
        n                   =   Floor[nf];
        nf                  =   nf-n;

        If[n<0,
            fi=arr[[1,1]]
            ,
            If[n<=lt-2,
                fi=(1-nf)*(1-nf)*arr[[n+1,n+1]]+2*nf*(1-nf)*arr[[n+1,n+2]]+nf*nf*arr[[n+2,n+2]]
                ,
                arr[[lt,lt]]
            ];
        ];

        matrix[[i,i]]       =   fi;
        Do[
            pj              =   p[[j]];
            mf              =   (lt-1)(pj-lpmin)/(lpmax-lpmin);
            m               =   Floor[mf];
            mf              =   mf-m;

            If[n<0,n=0];
            If[m<0,m=0];
            If[n>lt-2,n=lt-2];
            If[m>lt-2,m=lt-2];

            crossx          =   (1-nf)*(1-mf)*arr[[m+1,n+1]]+nf*(1-mf)*arr[[m+1,n+2]]+(1-nf)*mf*arr[[m+2,n+1]]+nf*mf*arr[[m+2,n+2]];
            matrix[[i,j]]   =   matrix[[j,i]]=crossx;
        ,
        {j,i+1,dim}
        ];
    ,
    {i,1,dim}
    ];

    pref*Exp[matrix]

], RuntimeOptions->"Speed", CompilationTarget->"C"];

(*----------------------------------------------------------------------------*)

fBENewtonItCFP=Compile[{{errNewton,_Real,0},{errMaxNewton,_Real,0},{dim,_Integer,0},{c1,_Real,0},{c2,_Real,0},{h,_Real,0},{xip1,_Real,0},{fi,_Real,1},{tq,_Real,1},{tq2,_Real,1},{Fi,_Real,1},{Derq1,_Real,2},{Derq2,_Real,2},{IM,_Real,2},{IM3,_Real,2},{AM,_Real,2},{gDH,_Real,0}},
Module[{i,fTip1, fEip1, xq, feqip1, Cscat, Ceqann, Cneqann, Jscat,Jann, deltafip1, NewtonError, NewtonTrapezoidal, NewtonEuler},

    (* Newton iteration of implicit Trapezoidal and Euler method. Scattering collision term in Fokker-Planck approximation *)


    fTip1                           =   fi; (*Predictor:*)
    fEip1                           =   fi; (*Predictor:*)
    xq                              =   Sqrt[xip1^2+c2*tq2];    (* c1 and c2 encode momentum coordinate dependence *)
    feqip1                          =   Exp[-xq];
    Ceqann                          =   (AM.(tq2 feqip1))*feqip1;
    Cscat                           =   ((c1) tq/xip1)(Derq1)+gDH*(1./xip1)((1./c2)xq(Derq2)+(tq+2.*xq/(c2*tq)+ tq/xq)*(Derq1)+IM3);

    (* Newton iteration of Trapezoidal method *)

    Jscat                           =   IM-(h/2.)*Cscat;

    Label[NewtonTrapezoidal];

        Cneqann                     =   AM.(tq2*fTip1);
        Jann                        =   (h/2.)AM Transpose[{fTip1}].{tq2};  (* Jacobian of Cann for off-diagonal terms*)

        For[i=1,i<=dim,i++,
            Jann[[i,i]]             =   (h/2.)(Cneqann[[i]]+AM[[i,i]]*tq2[[i]]*fTip1[[i]]);    (* Correct the diagonal terms*)
        ];

        deltafip1                   =   LinearSolve[Jscat+Jann,fTip1-fi-(h/2.)*(Fi+(Ceqann-(Cneqann*fTip1))+Cscat.fTip1)];
        fTip1                       =   fTip1-deltafip1;
        NewtonError                 =   Max[Abs[deltafip1/fTip1]];

        If[NewtonError>errNewton&&NewtonError<errMaxNewton,Goto[NewtonTrapezoidal]];

    (*  Newton iteration of Euler method    *)

    If[NewtonError<errMaxNewton,
        Jscat                       =   IM-(h)*Cscat;

        Label[NewtonEuler];

            Cneqann                 =   AM.(tq2*fEip1);
            Jann                    =   h AM Transpose[{fEip1}].{tq2};

            For[i=1,i<=dim,i++,
                Jann[[i,i]]         =   h(Cneqann[[i]]+AM[[i,i]]*tq2[[i]]*fEip1[[i]]);
            ];

            deltafip1               =   LinearSolve[Jscat+Jann,fEip1-fi-(h)*((Ceqann-(Cneqann*fEip1))+Cscat.fEip1)];
            fEip1                   =   fEip1-deltafip1;
            NewtonError             =   Max[Abs[deltafip1/fEip1]];

            If[NewtonError>errNewton&&NewtonError<errMaxNewton,Goto[NewtonEuler]];
    ];

    If[NewtonError<errMaxNewton,
        Cneqann                     =   (AM.(tq2*fTip1))*fTip1
        ,
        fEip1                       =   2fTip1; (* will lower the stepsize after return to fBE.wl*)
        Cneqann                     =   Ceqann
    ];

    {fTip1,fEip1,(Ceqann-Cneqann)+Cscat.fTip1}

], RuntimeOptions->"Speed", CompilationTarget->"C"];

(*----------------------------------------------------------------------------*)

fBENewtonItCel=Compile[{{errNewton,_Real,0},{errMaxNewton,_Real,0},{dim,_Integer},{c1,_Real},{c2,_Real},{h,_Real},{xip1,_Real},{fi,_Real,1},{tq,_Real,1},{tq2,_Real,1},{Fi,_Real,1},{Derq1,_Real,2},{IM,_Real,2},{AM,_Real,2},{SM,_Real,2}},
Module[{i,fTip1, fEip1, xq, feqip1, Cfree, Cfreescatter, Ceqann, Cneqann,Jann, M, deltafip1, NewtonError, NewtonTrapezoidal, NewtonEuler},

    (* Same as fBENewtonItCFP but with full Cell *)

    fTip1                           =   fi;
    fEip1                           =   fi;
    xq                              =   Sqrt[xip1^2+c2*tq2];
    feqip1                          =   Exp[-xq];
    Ceqann                          =   (AM.(tq2 feqip1))*feqip1;
    Cfree                           =   ((c1) tq/xip1)(Derq1);
    Cfreescatter                    =   Cfree+SM;

    (*Newton iteration of Trapezoidal method*)

    M                               =   IM-(h/2.)*(Cfreescatter);

    Label[NewtonTrapezoidal];

        Cneqann                     =   AM.(tq2*fTip1);
        Jann                        =   (h/2.)AM Transpose[{fTip1}].{tq2};

        For[i=1,i<=dim,i++,
            Jann[[i,i]]             =   (h/2.)(Cneqann[[i]]+AM[[i,i]]*tq2[[i]]*fTip1[[i]]);
        ];

        deltafip1                   =   LinearSolve[M+Jann,fTip1-fi-(h/2.)*(Fi+(Ceqann-(Cneqann*fTip1))+Cfreescatter.fTip1)];
        fTip1                       =   fTip1-deltafip1;
        NewtonError                 =   Max[Abs[deltafip1/fTip1]];

        If[NewtonError>errNewton&&NewtonError<errMaxNewton,Goto[NewtonTrapezoidal]];

    (*Newton iteration of Euler method*)

    If[NewtonError<errMaxNewton,
        M                           =   IM-(h)*(Cfreescatter);

        Label[NewtonEuler];

            Cneqann                 =   AM.(tq2*fEip1);
            Jann                    =   h AM Transpose[{fEip1}].{tq2};

            For[i=1,i<=dim,i++,
                Jann[[i,i]]         =   h(Cneqann[[i]]+AM[[i,i]]*tq2[[i]]*fEip1[[i]]);
            ];

            deltafip1               =   LinearSolve[M+Jann,fEip1-fi-(h)*((Ceqann-(Cneqann*fEip1))+Cfreescatter.fEip1)];
            fEip1                   =   fEip1-deltafip1;
            NewtonError             =   Max[Abs[deltafip1/fEip1]];

            If[NewtonError>errNewton&&NewtonError<errMaxNewton,Goto[NewtonEuler]];
    ];

    If[NewtonError<errMaxNewton,
        Cneqann                     =   (AM.(tq2*fTip1))*fTip1
        ,
        fEip1                       =   2fTip1;
        Cneqann                     =   Ceqann
    ];

    {fTip1,fEip1,(Ceqann-Cneqann)+Cfreescatter.fTip1}

], RuntimeOptions->"Speed", CompilationTarget->"C"];
