(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine solving the Boltzmann equation (nBE):

    Y'=-\[Lambda] (Y^2-Yeq^2)

Input as global function:

    svx[x]  -   thermally averaged annihilation cross section

Output:

    tYnBE   - table containing Y(x)
    Oh2nBE  - final value of relic abundance

Version: 15-01-2021

Remark: follows DarkSUSY implementaion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

nBE := Module[{xi,xf,h,hh,eps,safety,err,xip1,Yi,TrapYip1,EulYip1,
   Yeqip1,Yeqi,lambdai,lambdaip1,u,rho,c,cc,Again,End,tnBE},

    xi                  =   xmin+0.;
    xf                  =   xmax+0.;
    h                   =   nBEhinit+0.;
    eps                 =   nBEerr+0.;
    safety              =   nBEsafety+0.;

  (* Initialize *)

    Yeqi                =   Yeq[mDM,xi];
    Yi                  =   Yeqi;
    tnBE                =   {{xi,Yi}};
    lambdai             =   3.2262726*10^(18)isqrtgstar[mDM/xi]mDM/xi^2 svx[xi];

    While[xi<xf,

        hh              =   h;
        If[xi+hh>xf,hh=xf-xi];

        Label[Again];

            xip1        =   xi + hh;

            If[xip1==xi,Print["nBE.wl: error, stepsize overflow"];Goto[End]];

            Yeqip1      =   Yeq[mDM,xip1];
            lambdaip1   =   3.2262726*10^(18)isqrtgstar[mDM/xip1]mDM/xip1^2 svx[xip1];

            If[lambdaip1==0,Print["nBE.wl: svx zero at ", xip1];Goto[End]];

            u           =   hh lambdaip1;
            rho         =   lambdai/lambdaip1;
            c           =   2.*Yi+u((Yeqip1^2+rho Yeqi^2)-rho Yi^2);

            If[1.+u c<0.,hh=hh/2.;Goto[Again]];

            TrapYip1    =   c/(1.+Sqrt[1.+u c]);
            cc          =   4.*(Yi+u Yeqip1^2);
            EulYip1     =   cc/2./(1.+Sqrt[1.+u cc]);
            err         =   Abs[(TrapYip1-EulYip1)/(TrapYip1 eps)];

            If[err>1.,hh=Max[safety/Sqrt[err],0.1]hh;Goto[Again]];

        xi              =   xip1;
        Yi              =   TrapYip1;
        lambdai         =   lambdaip1;
        Yeqi            =   Yeqip1;
        AppendTo[tnBE,{xi,Yi}];

        If[err==0.,h=5.*hh,h=Min[safety/Sqrt[err],5.]*hh];

    ];

    Label[End];

    tYnBE               =   tnBE;
    Oh2nBE              =   mDM*Entropytoday/Rhocritical* Last[tYnBE][[2]];

    (* result for the relic density *)
    "Oh2nBE = " <> ToString[Oh2nBE]
];
