(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routines for computation of all the relevant thermal averages.

Version: 09-12-2020
Contact: Andrzej Hryczuk (NCBJ)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

(* thermal average of the annihilation cross section <sv> *)
TAsvx[x_] := Module[{f,limits,res},
(* Note: sv == sv_lab  as in Eq. (14) of 1706.07433 *)

   (* condition to avoid division by 0 when precision is not high enough
   to resolve BesselK *)
   f[s_] := If[x<xBexp,
     ((s-2 mDM^2)Sqrt[s - 4 mDM^2] BesselK1[Sqrt[s]/(mDM/x)])/
     (8 (mDM/x) mDM^4 BesselK2[x]^2) sv[s]
    ,
     1/(4mDM^2)((E^(-2 (-1 + Sqrt[st]) x) Sqrt[-1 +st] (-1 + 2 st) Sqrt[1/x]
     (-15 + 24 Sqrt[st] (-15 + 4 x) + 16 st (285 - 120 x + 32 x^2)))/
     (256 Sqrt[Pi] st^(5/4))) sv[s]/.st->s/(4mDM^2)
    ];

   (* if thresholds or resonances present include them explicitely in the integration limits *)
   If[!resonance=={}&&!threshold=={},Print["Warning: Resonance+Threshold currently not avaiable;
                                      proceeding with integrations without this information.
                                      The result might be inaccurate!"];
      ];
   If[!resonance=={},
     If[resonance[[1]]<4 mDM^2(1+20/x),
      limits=Flatten[{4 mDM^2,resonance,4 mDM^2(1+20/x)}];,
      limits=Flatten[{4 mDM^2,4 mDM^2(1+20/x),Min[4 mDM^2(1+30/x),2 resonance[[1]]]}];
       ];
     ];
   If[!threshold=={},
       limits=Flatten[{threshold,\[Infinity]}];
     ];
   If[resonance=={}&&threshold=={},limits={4 mDM^2,(1+20/x) 4 mDM^2};
     ];

   (* integration *)
    res=NIntegrate[f[ss],Flatten[{ss,limits}],
     WorkingPrecision -> MachinePrecision,
     MinRecursion -> 1,
     MaxRecursion -> 20,
     PrecisionGoal->pg,
     Method -> {"GlobalAdaptive", "SymbolicProcessing" -> False}];

    Return[res/.{0->minnumber,0.->minnumber,Indeterminate->minnumber}];

];

(* ---------------------------------------------------------------------------*)
(* thermal average of the annihilation p^2 weighted cross section , <sv>_2  *)
TAsv2x[x_] := Module[{f, ff, ep, tab, tab1, tab2, int, si, tmp,limits,accctime,accc},
(* Note: sv == sv_lab as in Eq. (30) of 1706.07433 *)

(* if thresholds or resonances present include them explicitly in the integration limits *)
If[!resonance=={}&&!thr=={},Print["Warning: Resonance+Threshold currently not avaiable;
                                   proceeding with integrations without this information.
                                   The result might be inaccurate!"];
   ];
   If[!resonance=={},
     If[resonance[[1]]<4 mDM^2(1+20/x),
      limits=Flatten[{1,resonance/(4 mDM^2),(1+20/x)}];,
      limits=Flatten[{1,(1+20/x),Min[(1+30/x),resonance[[1]]/4 mDM^2]}];
       ];
    ];
   If[!threshold=={},
       limits=Flatten[{threshold/(4mDM^2),threshold/(4mDM^2)+100/x}];
     ];
   If[resonance=={}&&threshold=={},limits={1 ,1+20/x};
     ];

     (* for low x use 2dim full integral, for high x expand to get 1dim *)
     If[x<xBexp,
      ff[s_?NumericQ] :=
       NIntegrate[(2 Sqrt[s] (2 s - 1) x^3 )/(3) ExpDbk2[x]^2
          Exp[-2 Sqrt[s] x ep+ 2 x] (Log[(Sqrt[s] ep - Sqrt[(s - 1) (ep^2 - 1)])/(Sqrt[s] ep + Sqrt[(s - 1) (ep^2 - 1)])]),
            {ep,1,1+20/x},
            Method -> {"GlobalAdaptive", "MaxErrorIncreases"->1000, "SymbolicProcessing" -> False},
            WorkingPrecision -> MachinePrecision,
            PrecisionGoal->pg,
            MaxRecursion -> 100
      ];

      (* integrand: two terms, one analytical and one numerical integral above *)
      f[s_] :=
        ((2 Sqrt[s](2s-1) x^3 )/(3BesselK[2,x]^2)(Sqrt[(-1+s)s]BesselK[2,2Sqrt[s]x])/(Sqrt[s] x)
        +ff[s]) sv[4 mDM^2 s];

     (* integration *)
     result=NIntegrate[f[ss], Flatten[{ss,limits}],
       WorkingPrecision -> MachinePrecision,
       MaxRecursion -> 100,
       MinRecursion -> 1,
       PrecisionGoal->pg,
       Method -> {"GlobalAdaptive", "MaxErrorIncreases"->1000, "SymbolicProcessing" -> False}];
    ,
      result = If[BesselK2[x]>0,
                NIntegrate[sv[4 mDM^2 ss]
                 (E^(-2 (-1 + Sqrt[ss]) x) Sqrt[(-1 +ss)/ss] (-1 + 2 ss)
                    (-225 - 88 Sqrt[ss] (-15 + 4 x) +
                    280 ss^(3/2) (-15 + 4 x) + ss (-3615 + 1920 x - 512 x^2) +
                    16 ss^2 (285 - 120 x + 32 x^2)))/(768 Sqrt[Pi] ss^(3/4)Sqrt[1/x]),
                 Flatten[{ss,limits}],
                 MaxRecursion -> 100,
                 MinRecursion -> 1,
                 PrecisionGoal->pg,
                 Method -> {"GlobalAdaptive", "MaxErrorIncreases"->1000, "SymbolicProcessing" -> False},
                 WorkingPrecision -> MachinePrecision
                ]
               ,minnumber];
    ];
    Return[result/.{0->minnumber,0.->minnumber,Indeterminate->minnumber}];

];

(* ---------------------------------------------------------------------------*)
(* thermal average of the non-relativistic annihilation cross section <sv> *)
TAsvxnonrel[x_] := Module[{vmin,vmax,Nv,dv,v,Maxf,v2,crossnonrel,res},
Nv=400;
vmin=10^(-4)Sqrt[2/x];
vmax=8.*Sqrt[2/x];
dv=(vmax-vmin)/(Nv-1.);
v=Table[vmin+(k-1.) dv,{k,Nv}];
v2=v^2;
Maxf=Exp[-x*v2/4];
crossnonrel=svnonrel[v];
res=(x^(3/2)/(2. Sqrt[Pi]))*dv*(Maxf.(v2*crossnonrel));
Return[Re[res]/.{0->minnumber,0.->minnumber,Indeterminate->minnumber}];
];

(* thermal average of the non-relativistic p^2 weighted cross section <sv>_2 *)
TAsv2xnonrel[x_] := Module[{vmin,vmax,Nv,dv,v,Maxf,v2,crossnonrel,res},
Nv=400;
vmin=10^(-4)Sqrt[2/x];
vmax=8.*Sqrt[2/x];
dv=(vmax-vmin)/(Nv-1.);
v=Table[vmin+(k-1.) dv,{k,Nv}];
v2=v^2;
Maxf=Exp[-x*v2/4];
crossnonrel=svnonrel[v];
res=(x^(3/2)/(4. Sqrt[Pi]))*dv*(Maxf.((1+(1/6)x v^2)*(v2*crossnonrel)));
Return[Re[res]/.{0->minnumber,0.->minnumber,Indeterminate->minnumber}];
];
